package cloudloc.client;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Calendar;
import java.util.LinkedList;
import java.util.List;

import com.amazonaws.AmazonClientException;
import com.amazonaws.AmazonServiceException;
import com.amazonaws.auth.BasicAWSCredentials;
import com.amazonaws.services.s3.AmazonS3;
import com.amazonaws.services.s3.AmazonS3Client;
import com.amazonaws.services.s3.model.Bucket;
import com.amazonaws.services.s3.model.PutObjectRequest;

/**
 * Class for a source uploader that uploads target files to S3 Run this class
 * "once" to prepare for the task.
 * 
 * @author f-ishikawa
 * 
 */
public class FileUploader {

	/**
	 * Initialize by creating a bucket in S3
	 * 
	 * @param bucketName
	 *            Name of the bucket to be created
	 */
	public static void createBucket(String bucketName) {
		AmazonS3 s3 = new AmazonS3Client(new BasicAWSCredentials(
				ClientConfig.ACCESS_KEY, ClientConfig.SEC_KEY));
		try {
			/* Do nothing if already exists */
			for (Bucket bucket : s3.listBuckets()) {
				if (bucket.getName().equals(bucketName)) {
					return;
				}
			}
			/* Create a bucket */
			System.out.println("Creating bucket in S3: " + bucketName + "\n");
			s3.createBucket(bucketName);
		} catch (AmazonServiceException e) {
			e.printStackTrace();
		} catch (AmazonClientException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Upload a file with a key into S3
	 * 
	 * @param key
	 * @param file
	 */
	public static void upload(String bucketName, String key, File file) {
		AmazonS3 s3 = new AmazonS3Client(new BasicAWSCredentials(
				ClientConfig.ACCESS_KEY, ClientConfig.SEC_KEY));
		try {
			System.out.println("Uploading a new object to S3 from a file\n");
			s3.putObject(new PutObjectRequest(bucketName, key, file));
		} catch (AmazonServiceException e) {
			e.printStackTrace();
		} catch (AmazonClientException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Put all files in the target directory into S3, using the relative path
	 * from the directory as a key for each file
	 * 
	 * @param bucketName Name of the bucket to which the files are uploaded
	 * @param dirPath
	 *            Path of the target directory
	 * @param reportFile Path of the file where the keys of the uploaded files are recorded           
	 * @return List contains all the keys for the files that were put in S3
	 */
	public static List<String> uploadAll(String bucketName, String dirPath, String reportFile) {
		/* Call a sub method */
		List<String> list = uploadAll_sub(bucketName, new File(dirPath), "");

		/* Write down all the keys put into S3 */
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(
					reportFile));
			for (String s : list) {
				bw.write(s);
				bw.append('\n');
			}
			bw.close();
		} catch (IOException e) {
			e.printStackTrace();
		}

		return list;
	}

	/**
	 * Sub method for uploadAll
	 */
	private static List<String> uploadAll_sub(String bucketName, File dir, String path) {
		LinkedList<String> list = new LinkedList<String>();
		for (File f : dir.listFiles()) {
			if (f.isDirectory()) {
				/* Upload recursively for an internal directory */
				list.addAll(uploadAll_sub(bucketName, f, path + "." + f.getName()));
			} else {
				/* Upload an internal (non-directory) file */
				/*
				 * The key is a relative path from the original directory, with
				 * the delimiter . (dot)
				 */
				String key = path + "." + f.getName();
				list.add(key);
				System.out.println("Uploading a file to S3: " + key + "\n");
				upload(bucketName, key, f);
			}
		}
		return list;
	}

	/**
	 * 
	 * @param args
	 *            No argument required
	 */
	public static void main(String[] args) {
		Calendar start = Calendar.getInstance();
		createBucket(ClientConfig.BUCKET_NAME);
		List<String> l = uploadAll(ClientConfig.BUCKET_NAME, ClientConfig.JAVA_SRC_PATH, ClientConfig.TASK_FILE);
		System.out.println(l.size());
		Calendar end = Calendar.getInstance();
		System.out.println("Upload started at " + start + " and ended at " + end);
	}

}
