package cloudloc.client;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Calendar;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

import javax.xml.namespace.QName;

import cloudloc.serviceaccess.generated.LocCountService;
import cloudloc.serviceaccess.generated.LocCountServicePortType;

import com.amazonaws.AmazonClientException;
import com.amazonaws.AmazonServiceException;
import com.amazonaws.auth.BasicAWSCredentials;
import com.amazonaws.services.ec2.AmazonEC2;
import com.amazonaws.services.ec2.AmazonEC2Client;
import com.amazonaws.services.ec2.model.DescribeInstancesRequest;
import com.amazonaws.services.ec2.model.DescribeInstancesResult;
import com.amazonaws.services.ec2.model.Instance;
import com.amazonaws.services.ec2.model.Reservation;
import com.amazonaws.services.ec2.model.RunInstancesRequest;
import com.amazonaws.services.ec2.model.RunInstancesResult;
import com.amazonaws.services.ec2.model.StopInstancesRequest;
import com.amazonaws.services.sqs.AmazonSQS;
import com.amazonaws.services.sqs.AmazonSQSClient;
import com.amazonaws.services.sqs.model.CreateQueueRequest;
import com.amazonaws.services.sqs.model.DeleteMessageRequest;
import com.amazonaws.services.sqs.model.DeleteQueueRequest;
import com.amazonaws.services.sqs.model.ListQueuesRequest;
import com.amazonaws.services.sqs.model.ListQueuesResult;
import com.amazonaws.services.sqs.model.Message;
import com.amazonaws.services.sqs.model.ReceiveMessageRequest;
import com.amazonaws.services.sqs.model.ReceiveMessageResult;
import com.amazonaws.services.sqs.model.SendMessageRequest;

/**
 * Class for a task controller that makes job requests, and outputs the gathered
 * results. Run this class to initiate the task, and then activate the
 * LocCountWorker VMs. Run SourceUploader just once to put target files to S3,
 * before running this class.
 * 
 * @author f-ishikawa
 * 
 */
public class TaskController {

	/* Num of files to process */
	private static int tasknum;

	/* max num of task (especially for test) */
	private int maxtasknum = -1;

	/* Information of running VM instances */
	private static List<String> instanceIDs = new LinkedList<String>();

	/*
	 * Queue names in Amazon SQS for keeping job requests and job results
	 */
	private String QUEUE_REQUEST_NAME = "request-queue";
	private String QUEUE_RESULT_NAME = "result-queue";

	/* URLs of the queues */
	private String requestQueueURL;
	private String resultQueueURL;

	/**
	 * Initialize by creating the queues to hold requests and results
	 */
	public void initialize() {
		AmazonSQS sqs = new AmazonSQSClient(new BasicAWSCredentials(
				ClientConfig.ACCESS_KEY, ClientConfig.SEC_KEY));
		try {
			/* Delete previous queues if remain undeleted */
			boolean flag = false;
			ListQueuesResult res;
			res = sqs.listQueues(new ListQueuesRequest(QUEUE_REQUEST_NAME));
			for (String url : res.getQueueUrls()) {
				sqs.deleteQueue(new DeleteQueueRequest(url));
				flag = true;
			}
			res = sqs.listQueues(new ListQueuesRequest(QUEUE_RESULT_NAME));
			for (String url : res.getQueueUrls()) {
				sqs.deleteQueue(new DeleteQueueRequest(url));
				flag = true;
			}
			if (flag) {
				/*
				 * Wait because the same name cannot be used for a while
				 */
				try {
					Thread.sleep(90000);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}

			/* Create the two queues and keep the URLs */
			System.out
					.println("Creating new SQS queues to hold requests and results");
			requestQueueURL = sqs.createQueue(
					new CreateQueueRequest(QUEUE_REQUEST_NAME)).getQueueUrl();
			resultQueueURL = sqs.createQueue(
					new CreateQueueRequest(QUEUE_RESULT_NAME)).getQueueUrl();
		} catch (AmazonServiceException e) {
			e.printStackTrace();
		} catch (AmazonClientException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Read the task file, and put all the file IDs to be processed into the
	 * request queue
	 */
	public void requestAll() {
		tasknum = 0;
		AmazonSQS sqs = new AmazonSQSClient(new BasicAWSCredentials(
				ClientConfig.ACCESS_KEY, ClientConfig.SEC_KEY));
		System.out
				.println("Adding all the job request in the task file to the request queue");
		try {
			/* Read the task file */
			BufferedReader br = new BufferedReader(new FileReader(
					ClientConfig.TASK_FILE));
			String fileID;
			while (br.ready()) {
				fileID = br.readLine();
				System.out
						.println("Adding a job request to the request queue: "
								+ fileID);
				try {
					sqs.sendMessage(new SendMessageRequest(requestQueueURL,
							fileID));
				} catch (AmazonServiceException e) {
					e.printStackTrace();
					System.err.println("Could not send a request on " + fileID);
					continue;
				} catch (AmazonClientException e) {
					e.printStackTrace();
					System.err.println("Could not send a request on " + fileID);
					continue;
				}
				tasknum++;
				if (maxtasknum > 0 && maxtasknum == tasknum) {
					break;
				}
			}
			br.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Activate worker VM instances
	 */
	public boolean runEC2instances(int instanceNum, String instanceType) {
		/* Run VM instances */
		AmazonEC2 ec2 = new AmazonEC2Client(new BasicAWSCredentials(
				ClientConfig.ACCESS_KEY, ClientConfig.SEC_KEY));
		try {
			System.out.println("Running " + instanceNum
					+ " worker VM instances");
			RunInstancesRequest req = new RunInstancesRequest(
					ClientConfig.AMI_ID, instanceNum, instanceNum);
			req.setInstanceType(instanceType);
			req.withSecurityGroupIds(ClientConfig.SECURITY_GROUP_ID);
			RunInstancesResult res = ec2.runInstances(req);
			for (Instance i : res.getReservation().getInstances()) {
				instanceIDs.add(i.getInstanceId());
			}
		} catch (AmazonServiceException e) {
			e.printStackTrace();
			return false;
		} catch (AmazonClientException e) {
			e.printStackTrace();
			return false;
		}
		return true;
	}

	public boolean startLOCservices() {
		AmazonEC2 ec2 = new AmazonEC2Client(new BasicAWSCredentials(
				ClientConfig.ACCESS_KEY, ClientConfig.SEC_KEY));
		int failcount = 0;
		boolean allrunning = false;
		List<String> addresses = new LinkedList<String>();
		while (true) {
			/* Wait for the worker VMs to run */
			try {
				DescribeInstancesResult res = ec2
						.describeInstances(new DescribeInstancesRequest()
								.withInstanceIds(instanceIDs));
				allrunning = true;
				for (Reservation r : res.getReservations()) {
					for (Instance i : r.getInstances()) {
						if (!i.getState().getName().equals("running")) {
							System.out.println("Instance " + i.getInstanceId()
									+ " is not yet running");
							allrunning = false;
						}
					}
				}
				if (allrunning) {
					System.out
							.println("Confirmed all the instances are running now");
					for (Reservation r : res.getReservations()) {
						for (Instance i : r.getInstances()) {
							addresses.add(i.getPublicDnsName());
						}
					}
					break;
				} else {
					if (failcount == 4) {
						return false;
					}
					failcount++;
					try {
						Thread.sleep(30000);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
			} catch (AmazonServiceException e1) {
				e1.printStackTrace();
			} catch (AmazonClientException e1) {
				e1.printStackTrace();
			}
		}

		/* Invoke LOC service on the VMs */
		LocCountServicePortType locservice;
		for (String address : addresses) {
			/* Retry a few times */
			for (int i = 0; i < 3; i++) {
				String url = "http://" + address
						+ ":8080/axis2/services/LocCountService?wsdl";
				System.out.println("Invoking a LOC service at " + url);
				try {
					locservice = new LocCountService(new URL(url), new QName(
							"http://service.cloudloc/", "LocCountService"))
							.getLocCountServiceHttpSoap11Endpoint();
					locservice.countAll(requestQueueURL, resultQueueURL,
							ClientConfig.BUCKET_NAME);
					break;
				} catch (MalformedURLException e) {
					e.printStackTrace();
					return false;
				} catch (RuntimeException e) {
					e.printStackTrace();
					try {
						Thread.sleep(30000);
					} catch (InterruptedException e1) {
						e1.printStackTrace();
					}
					if (i == 2) {
						return false;
					}
				}
			}
		}

		return true;
	}

	/**
	 * Complete the task by reading the result queue
	 */
	public void close() {
		AmazonSQS sqs = new AmazonSQSClient(new BasicAWSCredentials(
				ClientConfig.ACCESS_KEY, ClientConfig.SEC_KEY));

		HashMap<String, HashMap<String, String>> results = new HashMap<String, HashMap<String, String>>();

		try {
			FileWriter fw = new FileWriter("result.txt");
			/*
			 * Repeat until the result queue is kept empty for a certain period
			 * or all the jobs are finished
			 */
			int failcount = 0;
			while (true) {
				System.out.println("Receiving messages from the result queue");
				ReceiveMessageResult res;
				try {
					res = sqs.receiveMessage(new ReceiveMessageRequest(
							resultQueueURL));
				} catch (AmazonServiceException e1) {
					e1.printStackTrace();
					break;
				} catch (AmazonClientException e1) {
					e1.printStackTrace();
					break;
				}
				List<Message> messages = res.getMessages();
				System.out.println("Received " + messages.size()
						+ " messages from the result queue");

				if (messages.isEmpty()) {
					System.out
							.println("Consumed all the messages in the result queue");
					if (results.size() == tasknum || failcount == 9) {
						fw.write("Total: " + results.size() + " Files");
						fw.close();
						break;
					}
					failcount++;
					fw.close();
					try {
						System.out
								.println("No result available from the result queue, sleeping (currently have "
										+ results.size()
										+ "/"
										+ tasknum
										+ " results)");
						Thread.sleep(60000);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
					fw = new FileWriter("result.txt", true);
				} else {
					failcount = 0;
					/*
					 * Each message in the result queue contains a key in S3,
					 * which refers to a result from one worker
					 */
					for (Message message : messages) {
						fw.write(message.getBody() + "\n");

						StringTokenizer st = new StringTokenizer(
								message.getBody());
						String host = st.nextToken();
						String fileID = st.nextToken();
						String num = st.nextToken();
						if (!results.containsKey(fileID)) {
							results.put(fileID, new HashMap<String, String>());
						} else {
							for (String n : results.get(fileID).values()) {
								if (!n.equals(num)) {
									System.err
											.println("Report: different result on "
													+ fileID);
								}
							}
						}
						results.get(fileID).put(host, num);

						System.out
								.println("Deleting a received message from the result queue");
						try {
							sqs.deleteMessage(new DeleteMessageRequest(
									resultQueueURL, message.getReceiptHandle()));
						} catch (AmazonServiceException e) {
							e.printStackTrace();
						} catch (AmazonClientException e) {
							e.printStackTrace();
						}
					}
				}
			}
		} catch (IOException e1) {
			e1.printStackTrace();
		}
		System.out.println("Finished with " + results.size() + "/" + tasknum
				+ " results)");
	}

	/**
	 * Delete the queues and stop VM instances
	 */
	public void finalize() {
		AmazonSQS sqs = new AmazonSQSClient(new BasicAWSCredentials(
				ClientConfig.ACCESS_KEY, ClientConfig.SEC_KEY));
		/* Delete the two queues */
		System.out.println("Deleting the request and result queues");
		try {
			sqs.deleteQueue(new DeleteQueueRequest(requestQueueURL));
			sqs.deleteQueue(new DeleteQueueRequest(resultQueueURL));
		} catch (AmazonServiceException e) {
			e.printStackTrace();
		} catch (AmazonClientException e) {
			e.printStackTrace();
		}

		/* Stop the instances */
		System.out.println("Stopping worker VM instances");
		AmazonEC2 ec2 = new AmazonEC2Client(new BasicAWSCredentials(
				ClientConfig.ACCESS_KEY, ClientConfig.SEC_KEY));
		try {
			ec2.stopInstances(new StopInstancesRequest(instanceIDs));
		} catch (AmazonServiceException e) {
			e.printStackTrace();
		} catch (AmazonClientException e) {
			e.printStackTrace();
		}
	}

	/**
	 * 
	 * @param args
	 *            The first argument denotes time (min) to wait for the workers,
	 *            if there is
	 */
	public static void main(String[] args) {
		TaskController tc = new TaskController();

		int num = 2;
		String type = "t1.micro";
		System.out
				.println("[Usage] first argument: num of instances (default 2)");
		System.out
				.println("        second argument: instance type (default t1.micro)");
		System.out
				.println("        third argument: max number of tasks (default no limitation)");
		if (args.length > 0) {
			num = Integer.parseInt(args[0]);
		}
		if (args.length > 1) {
			type = args[1];
		}
		if (args.length > 2) {
			tc.maxtasknum = Integer.parseInt(args[2]);
		}

		Calendar c1 = Calendar.getInstance();
		tc.initialize();
		tc.requestAll();
		Calendar c2 = Calendar.getInstance();

		boolean result = tc.runEC2instances(num, type);
		if (!result) {
			tc.finalize();
			return;
		}
		result = tc.startLOCservices();
		if (!result) {
			tc.finalize();
			return;
		}

		tc.close();
		tc.finalize();
		Calendar c3 = Calendar.getInstance();

		System.out.println("Task started at " + c1.get(Calendar.HOUR_OF_DAY)
				+ ":" + c1.get(Calendar.MINUTE) + ", all requests sent out at "
				+ c2.get(Calendar.HOUR_OF_DAY) + ":" + c2.get(Calendar.MINUTE)
				+ " and finished at " + c3.get(Calendar.HOUR_OF_DAY) + ":"
				+ c3.get(Calendar.MINUTE));
	}

}
