package cloudloc.service;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.List;

import com.amazonaws.AmazonClientException;
import com.amazonaws.AmazonServiceException;
import com.amazonaws.auth.BasicAWSCredentials;
import com.amazonaws.services.s3.AmazonS3;
import com.amazonaws.services.s3.AmazonS3Client;
import com.amazonaws.services.s3.model.GetObjectRequest;
import com.amazonaws.services.s3.model.S3Object;
import com.amazonaws.services.sqs.AmazonSQS;
import com.amazonaws.services.sqs.AmazonSQSClient;
import com.amazonaws.services.sqs.model.DeleteMessageRequest;
import com.amazonaws.services.sqs.model.Message;
import com.amazonaws.services.sqs.model.ReceiveMessageRequest;
import com.amazonaws.services.sqs.model.SendMessageRequest;

/**
 * Class for LOC count service. Deploy this POJO (Plain Old Java Object) as a
 * service by putting the class file with META-INF/services.xml into the
 * "services" directory inside Axis2 installation.
 * 
 * @author f-ishikawa
 * 
 */
public class LocCountService {

	/* Access keys */
	private static String ACCESS_KEY = "XXXXXXXXXXXXXXX";
	private static String SEC_KEY = "XXXXXXXXXXXXXXX";

	/**
	 * Count LOC in texts read from input stream
	 * 
	 * @param inputstream
	 * @return
	 */
	private int count(InputStream inputstream) {
		BufferedReader br = new BufferedReader(new InputStreamReader(
				inputstream));
		int count = 0;
		try {
			while (br.ready()) {
				br.readLine();
				count++;
			}
		} catch (IOException e) {
			e.printStackTrace();
			System.err.println("\nLine count failed");
			return 0;
		}
		return count;
	}

	/**
	 * Count LOC in texts read from a file in S3
	 * 
	 * @param objectKey
	 *            Key to get a file from S3
	 * @return
	 */
	private int count(String bucketName, String objectKey) {
		AmazonS3 s3 = new AmazonS3Client(new BasicAWSCredentials(ACCESS_KEY,
				SEC_KEY));
		System.out.println("Downloading a file from Amazon S3: " + objectKey);
		try {
			S3Object obj = s3.getObject(new GetObjectRequest(bucketName,
					objectKey));
			System.out.println("Counting LOC of the file");
			return count(obj.getObjectContent());
		} catch (AmazonServiceException e) {
			e.printStackTrace();
		} catch (AmazonClientException e) {
			e.printStackTrace();
		}
		System.out.println("Failured to download a file from Amazon S3");
		return 0;
	}

	/**
	 * Read the request queue, execute and put the result into the result queue.
	 * Public method published as a service operation, which does not block the
	 * requester.
	 * 
	 * @param requestQueueURL
	 *            URL of the request queue (target object keys in S3 are put by
	 *            the requester(s))
	 * @param resultQueueURL
	 *            URL of the result queue (result reports are put by this
	 *            method)
	 * @param bucketName
	 *            Name of the S3 bucket from which target files are read
	 */
	public void countAll(String requestQueueURL, String resultQueueURL,
			String bucketName) {
		/* Execute asynchronously */
		new Thread(new Run(this, requestQueueURL, resultQueueURL, bucketName))
				.start();
	}

	void countAll_actual(String requestQueueURL, String resultQueueURL,
			String bucketName) {
		/* Identifier of the worker running this method (IP address) */
		String workerID;
		try {
			workerID = InetAddress.getLocalHost().getHostAddress();
		} catch (UnknownHostException e1) {
			workerID = "UnknownHost";
			e1.printStackTrace();
		}

		AmazonSQS sqs = new AmazonSQSClient(new BasicAWSCredentials(ACCESS_KEY,
				SEC_KEY));

		/* Repeat until the request queue is kept empty for a certain period */
		int failcount = 0;
		int count = 0;
		while (true) {
			System.out.println("Receiving messages from the request queue.");
			List<Message> messages;
			try {
				messages = sqs.receiveMessage(
						new ReceiveMessageRequest(requestQueueURL))
						.getMessages();
			} catch (AmazonServiceException e) {
				e.printStackTrace();
				messages = null;
			} catch (AmazonClientException e) {
				e.printStackTrace();
				messages = null;
			}

			/* If fail */
			if (messages == null || messages.isEmpty()) {
				System.out.println("No request to process");
				/* Stops if keep failing or have no request for a few minutes */
				if (failcount == 9) {
					break;
				}
				try {
					Thread.sleep(30000);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
				failcount++;
			} else {
				failcount = 0;
				for (Message message : messages) {
					count++;
					/*
					 * Measure LOC, report the result to the result queue, and
					 * delete the request from the request queue
					 */
					try {
						sqs.sendMessage(new SendMessageRequest(resultQueueURL,
								workerID + " " + message.getBody() + " "
										+ count(bucketName, message.getBody())));
					} catch (AmazonServiceException e) {
						e.printStackTrace();
						continue;
					} catch (AmazonClientException e) {
						e.printStackTrace();
						continue;
					}
					try {
						sqs.deleteMessage(new DeleteMessageRequest(
								requestQueueURL, message.getReceiptHandle()));
					} catch (AmazonServiceException e) {
						e.printStackTrace();
					} catch (AmazonClientException e) {
						e.printStackTrace();
					}
				}
			}
		}
		System.out.println("Processed " + count + " requests");
	}

}

class Run implements Runnable {

	private LocCountService service;
	private String requestQueueURL;
	private String resultQueueURL;
	private String bucketName;

	Run(LocCountService service, String requestQueueURL, String resultQueueURL,
			String bucketName) {
		this.service = service;
		this.requestQueueURL = requestQueueURL;
		this.resultQueueURL = resultQueueURL;
		this.bucketName = bucketName;
	}

	@Override
	public void run() {
		service.countAll_actual(requestQueueURL, resultQueueURL, bucketName);
	}

}